# -*- coding: utf-8 -*-
import skimage.io
import numpy as np
import pycuda.autoinit
import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
import skcuda.linalg as linalg
import skcuda.misc as misc
from pycuda.elementwise import ElementwiseKernel
import pycuda.cumath as cumath
import sys, getopt
import time

# El step_1 realiza la multiplicaci�n elemento a elemento de una banda con la pancrom�tica
def step_1(color_matrix, image_matrix):
    #La funciÃ³n linalg.mulitply realiza la multiplicaciÃ³n elemento a elemento entre dos matrices
    matrix_sal = linalg.multiply(color_matrix, image_matrix)
    return matrix_sal

# El step_2 determina el valor maximo y minimo de la banda resultante de la transformada de brovey
def step_2(matrix_1):
    mat_max = np.amax(matrix_1)
    mat_min = np.amin(matrix_1)
    return mat_max, mat_min


def step_3(matrix_1, matrix_color, mat_max, mat_min):
    #La funciÃ³n lin_comb invoca el kernel elementwise establecido para llevar a cabo el ajuste por rescale global min
    lin_comb(mat_min, matrix_1, mat_max, matrix_color)
    return matrix_color.get()


# Kernel establecido para llevar a cabo el mÃ©todo de rescale global min
lin_comb = ElementwiseKernel(
        "float a, float *x, float b, float *z",
        "z[i] = ((x[i]-a)*255)/(b-a)",
        "linear_combination")


def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    #Verifica que ambas imagenes cumplan con las condiciones
    end = 0
    start = 0
    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')

    # Convierte a float32 y separa las bandas RGB de la multispectral
    multispectral = multispectral.astype(np.float32)
    r = multispectral[:,:,0].astype(np.float32)
    g = multispectral[:,:,1].astype(np.float32)
    b = multispectral[:,:,2].astype(np.float32)
    # Convierte la pancromatica a float32
    panchromatic = panchromatic.astype(np.float32)
    # Suma las bandas de la multispectral
    msuma = r+g+b
    start=time.time()

    r_gpu = gpuarray.to_gpu(r)
    g_gpu = gpuarray.to_gpu(g)
    b_gpu = gpuarray.to_gpu(b)
    panchromatic_gpu = gpuarray.to_gpu(panchromatic)

    linalg.init()

    m33_gpu = step_1(r_gpu, panchromatic_gpu)
    m44_gpu = step_1(g_gpu, panchromatic_gpu)
    m55_gpu = step_1(b_gpu, panchromatic_gpu)

    Amax, Amin = step_2(m33_gpu.get())
    br_gpu = gpuarray.empty_like(r_gpu)
    br_host = step_3(m33_gpu, br_gpu, Amax, Amin)

    Amax, Amin = step_2(m44_gpu.get())
    bg_gpu = gpuarray.empty_like(g_gpu)
    bg_host = step_3(m44_gpu, bg_gpu, Amax, Amin)

    Amax, Amin = step_2(m55_gpu.get())
    bb_gpu = gpuarray.empty_like(b_gpu)
    bb_host = step_3(m55_gpu, bb_gpu, Amax, Amin)

    end = time.time()

    brr = br_host.astype(np.uint8)
    bgg = bg_host.astype(np.uint8)
    bbb = bb_host.astype(np.uint8)

    fusioned_image = np.stack((brr, bgg, bbb),axis=2)

    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/multiplicativegpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('multiplicativegpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image



# -*- coding: utf-8 -*-
import skimage.io
import numpy as np
import cv2
import sys, getopt
import time


# El step_1 realiza la la división de la banda entre la suma de las bandas
def step_1(matrix_1, matrix_color, msuma_matrix):
    for m in range(matrix_1.shape[0]):
        for n in range(matrix_1.shape[0]):
            if (msuma_matrix[m,n] != 0):
                matrix_1[m,n] = (3*matrix_color[m,n])/msuma_matrix[m,n]
            else:
                matrix_1[m,n] = msuma_matrix[m,n]
    return matrix_1

# El step_2 realiza la multiplicación posición a posición del step_1 por la pancromatica
def step_2(matrix_1, matrix_2, matrix_image_pan):
    for m in range(matrix_2.shape[0]):
        for n in range(matrix_2.shape[0]):
            matrix_2[m,n] = matrix_1[m,n]*matrix_image_pan[m,n]
    return matrix_2

# El step_3 determina el valor maximo y minimo de la banda resultante de la transformada de brovey
def step_3(matrix_1):
    mat_max = np.amax(matrix_1)
    mat_min = np.amin(matrix_1)
    return mat_max, mat_min

# El step_4 realiza un ajuste de riqueza espectral llamado rescale global min
def step_4(matrix_1, matrix_color, mat_max, mat_min):
    for m in range(matrix_color.shape[0]):
        for n in range(matrix_color.shape[0]):
            matrix_color[m,n] = (((matrix_1[m,n]-mat_min)*255)/(mat_max-mat_min))
    return matrix_color


def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    end = 0
    start = 0
    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')

    # Convierte a float32 y separa las bandas RGB de la multispectral
    m1 = multispectral.astype(np.float32)
    r = multispectral[:,:,0]
    r1 = r.astype(np.float32)
    g = multispectral[:,:,1]
    g1 = g.astype(np.float32)
    b = multispectral[:,:,2]
    b1 = b.astype(np.float32)
    # Convierte la pancromatica a float32
    p1 = panchromatic.astype(np.float32)
    msuma = r1+g1+b1
    start=time.time()
    m11 = np.zeros_like(r1)
    m11 = step_1(m11, r1, msuma)
    m22 = np.zeros_like(g1)
    m22 = step_2(m11, m22, p1)
    m33 = np.zeros_like(b1)
    m33 = step_1(m33, b1, msuma)
    m44 = np.zeros_like(b1)
    m44 = step_2(m33, m44, p1)
    m55 = np.zeros_like(g1)
    m55 = step_1(m55, g1, msuma)
    m66 = np.zeros_like(g1)
    m66 = step_2(m55, m66, p1)
    Amax, Amin = step_3(m22)
    rr = np.zeros_like(r1)
    rr = step_4(m22, rr, Amax, Amin)
    Amax, Amin = step_3(m66)
    gg = np.zeros_like(g1)
    gg = step_4(m66, gg, Amax, Amin)
    Amax, Amin = step_3(m44)
    bb = np.zeros_like(b1)
    bb = step_4(m44, bb, Amax, Amin)
    end = time.time()
    rrr = rr.astype(np.uint8)
    ggg = gg.astype(np.uint8)
    bbb = bb.astype(np.uint8)
    # Combina las bandas resultantes
    fusioned_image = np.stack((rrr, ggg, bbb),axis=2)
    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/broveycpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('broveycpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image



#-*- coding: utf-8 -*-
import skimage.io
from skimage.color import rgb2hsv, hsv2rgb
import numpy as np
import pycuda.autoinit
import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
from pycuda.elementwise import ElementwiseKernel
import cupy as cp
from cupyx.scipy.ndimage import filters
import sys
import time


# Kernel establecido para realiza un ajuste cuando el valor de un pixel es menor a 0
adjustment_values = ElementwiseKernel(
        "float *x, float *z",
        "if(x[i] < 0){z[i] = 0.0;}else{z[i] = x[i];}",
        "adjust_value")


def hist_match(source, template):
    oldshape = source.shape
    source = source.ravel()
    template = template.ravel()

    # get the set of unique pixel values and their corresponding indices and
    # counts
    s_values, bin_idx, s_counts = np.unique(source, return_inverse=True,
                                            return_counts=True)
    t_values, t_counts = np.unique(template, return_counts=True)

    # take the cumsum of the counts and normalize by the number of pixels to
    # get the empirical cumulative distribution functions for the source and
    # template images (maps pixel value --> quantile)
    s_quantiles = np.cumsum(s_counts).astype(np.float64)
    s_quantiles /= s_quantiles[-1]
    t_quantiles = np.cumsum(t_counts).astype(np.float64)
    t_quantiles /= t_quantiles[-1]

    # interpolate linearly to find the pixel values in the template image
    # that correspond most closely to the quantiles in the source image
    interp_t_values = np.interp(s_quantiles, t_quantiles, t_values)

    return interp_t_values[bin_idx].reshape(oldshape)

def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    end = 0
    start = 0
    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')


    start=time.time()

    hsv = rgb2hsv(multispectral)

    val = hsv[:,:,2]
    sat = hsv[:,:,1]
    mat = hsv[:,:,0]

    pani = hist_match(panchromatic,val)

    s = np.array([[1/256, 1/64, 3/128, 1/64, 1/256],[1/64, 1/16, 3/32, 1/16, 1/64],[3/128, 3/32, 9/64, 3/32, 3/128],
                        [1/64, 1/16, 3/32, 1/16, 1/64],[1/256, 1/64, 3/128, 1/64, 1/256]])

    s_gpu = cp.array(s)
    p_gpu = cp.array(pani)

    I1_gpu = filters.correlate(p_gpu, s_gpu, mode='constant')

    s1 = np.array([[1/256, 0, 1/64, 0, 3/128, 0, 1/64, 0, 1/256],[0, 0, 0, 0, 0, 0, 0, 0, 0],[1/64, 0, 1/16, 0, 3/32, 0, 1/16, 0, 1/64],[0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [3/128, 0, 3/32, 0, 9/64, 0, 3/32, 0, 3/128],[0, 0, 0, 0, 0, 0, 0, 0, 0],[1/64, 0, 1/16, 0, 3/32, 0, 1/16, 0, 1/64], [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [1/256, 0, 1/64, 0, 3/128, 0, 1/64, 0, 1/256]])

    s1_gpu = cp.array(s1)
    I2_gpu = filters.correlate(I1_gpu, s1_gpu, mode='constant')

    W1=(pani-I1_gpu.get())
    W1_gpu = gpuarray.to_gpu(W1)
    W1_gpu_new = gpuarray.empty_like(W1_gpu)
    adjustment_values(W1_gpu,W1_gpu_new)
    W1 = W1_gpu_new.get().astype(np.uint8)

    W2=(I1_gpu.get()-I2_gpu.get())
    W2_gpu = gpuarray.to_gpu(W2)
    W2_gpu_new = gpuarray.empty_like(W2_gpu)
    adjustment_values(W2_gpu,W2_gpu_new)
    W2 = W2_gpu_new.get().astype(np.uint8)

    nint=(panchromatic+W1+W2).astype(np.uint8)
    end = time.time()
    n_hsv1 = np.stack((mat, sat, nint),axis=2)
    n_hsv = n_hsv1
    fusioned_image = hsv2rgb(n_hsv).astype(np.uint8)
    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/atrousgpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('atrousgpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image



# -*- coding: utf-8 -*-
from sallfus.fusion import broveyCPU
from sallfus.fusion import broveyGPU
from sallfus.fusion import multiplicativeCPU
from sallfus.fusion import multiplicativeGPU
from sallfus.fusion import atrousCPU
from sallfus.fusion import atrousGPU
from sallfus.fusion import pcaCPU
from sallfus.fusion import pcaGPU
from sallfus.measures import mse, rmse, correlation_coeff, bias
import sys

def generated_method(method):
    if(method == 'brovey'):
        array_method = [broveyCPU, broveyGPU]
    elif(method == 'multiplicative'):
        array_method = [multiplicativeCPU, multiplicativeGPU]
    elif(method == 'pca'):
        array_method = [pcaCPU, pcaGPU]
    elif(method == 'atrous'):
        array_method = [atrousCPU, atrousGPU]
    else:
         sys.exit('The method is invalid')
    return array_method


def time_comparison(multispectral, panchromatic, method):
    results = dict()
    array_types = ['cpu','gpu']
    if(type(method) != str):
        sys.exit('Methods must be a string')
    method_array = generated_method(method)
    for i in range(len(method_array)):
        result_temp = method_array[i].fusion_images(multispectral, panchromatic)
        results.update({method+'_'+array_types[i] : result_temp['time']})
    return results


def generated_measures(measures):
    array_measures = []
    for i in range(len(measures)):
        if(measures[i] == 'mse'):
            array_measures.append(mse)
        elif(measures[i] == 'rmse'):
            array_measures.append(rmse)
        elif(measures[i] == 'cc'):
            array_measures.append(correlation_coeff)
        elif(measures[i] == 'bias'):
            array_measures.append(bias)
        else:
             sys.exit('The measure is invalid')
    return array_measures

def measures_comparison(fusioned, original, measures):
    results = dict()
    results_temp_dict = dict()
    real_measures = generated_measures(measures)
    for i in range(len(real_measures)):
        result_temp = real_measures[i](fusioned, original)
        results_temp_dict.update({'R':result_temp[0], 'G':result_temp[1], 'B':result_temp[2]})
        results.update({measures[i] : results_temp_dict})
        results_temp_dict = dict() 
    return results

def generated_methods(methods):
    array_methods = []
    array_names = []
    for i in range(len(methods)):
        if(methods[i] == 'brovey'):
            array_methods.append(broveyCPU)
            array_methods.append(broveyGPU)
            array_names.append('broveyCPU')
            array_names.append('broveyGPU')
        elif(methods[i] == 'multiplicative'):
            array_methods.append(multiplicativeCPU)
            array_methods.append(multiplicativeGPU)
            array_names.append('multiplicativeCPU')
            array_names.append('multiplicativeGPU')
        elif(methods[i] == 'pca'):
            array_methods.append(pcaCPU)
            array_methods.append(pcaGPU)
            array_names.append('pcaCPU')
            array_names.append('pcaGPU')
        elif(methods[i] == 'atrous'):
            array_methods.append(atrousCPU)
            array_methods.append(atrousGPU)
            array_names.append('atrousCPU')
            array_names.append('atrousGPU')
        else:
             sys.exit('The methods are invalid')
    return array_methods, array_names

def time_comparison_multiple(multispectral, panchromatic, methods):
    results = dict()
    array_types = ['cpu','gpu']
    methods_array, names_array = generated_methods(methods)
    for i in range(len(methods_array)):
        result_temp = methods_array[i].fusion_images(multispectral, panchromatic)
        results.update({names_array[i] : result_temp['time']})
    return results



# -*- coding: utf-8 -*-
import numpy as np

def check_images(fusioned, original):
    assert len(fusioned) == len(original), "Supplied images have different sizes " + \
    str(fusioned.shape) + " and " + str(original.shape)
    if(len(fusioned.shape) == len(original.shape)):
        estado = 'mtom'
        if(len(fusioned.shape) == 2):
            fusioned = fusioned[:,:,np.newaxis]
            original = original[:,:,np.newaxis]
        else:
            assert fusioned.shape[2] == original.shape[2], "Supplied images have different number of bands "
    else:
        estado = 'mtop'
    return estado, fusioned, original

def mse(fusioned, original):
    """calculates mean squared error (mse).
    :param GT: first (original) input image.
    :param P: second (deformed) input image.
    :returns:  float -- mse value.
    """
    array_mse = []
    mode, fusioned, original = check_images(fusioned, original)
    if(mode == 'mtom'):
        for i in range(fusioned.shape[2]):
            aux_val = np.mean((fusioned[:,:,i].astype(np.float64)-original[:,:,i].astype(np.float64))**2)
            array_mse.append(aux_val)
    else:
        for i in range(fusioned.shape[2]):
            aux_val = np.mean((fusioned[:,:,i].astype(np.float64)-original.astype(np.float64))**2)
            array_mse.append(aux_val)

    return np.array(array_mse)


def rmse(fusioned, original):
    return np.sqrt(mse(fusioned, original))


def bias(fusioned, original):
    array_bias = []
    mode, fusioned, original = check_images(fusioned, original)
    if(mode == 'mtom'):
        for i in range(fusioned.shape[2]):
            aux_val = 1 - ((np.mean(fusioned[:,:,i].astype(np.float64)))/ (np.mean(original[:,:,i].astype(np.float64))))
            array_bias.append(aux_val)
    else:
        for i in range(fusioned.shape[2]):
            aux_val = 1 - ((np.mean(fusioned[:,:,i].astype(np.float64)))/ (np.mean(original.astype(np.float64))))
            array_bias.append(aux_val)
    return array_bias


def correlation_coeff(fusioned, original):
    array_corrcoef = []
    mode, fusioned, original = check_images(fusioned, original)
    if(mode == 'mtom'):
        for i in range(fusioned.shape[2]):
            aux_val = np.corrcoef(fusioned[:,:,i].astype(np.float64).flat, original[:,:,i].astype(np.float64).flat)
            array_corrcoef.append(aux_val[0][1])
    else:
        for i in range(fusioned.shape[2]):
            aux_val = np.corrcoef(fusioned[:,:,i].astype(np.float64).flat, original.astype(np.float64).flat)
            array_corrcoef.append(aux_val[0][1])
    return array_corrcoef



from sallfus.fusion import broveyCPU
from sallfus.fusion import broveyGPU
from sallfus.fusion import multiplicativeCPU
from sallfus.fusion import multiplicativeGPU
from sallfus.fusion import pcaCPU
from sallfus.fusion import pcaGPU
from sallfus.fusion import atrousCPU
from sallfus.fusion import atrousGPU
from sallfus import comparison
from sallfus import measures



# -*- coding: utf-8 -*-
import skimage.io
import pycuda.autoinit
import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
import numpy as np
import skcuda.linalg as linalg
from pycuda.elementwise import ElementwiseKernel
import sys
import time


# El step_1 realiza la la divisiÃ³n de la banda entre la suma de las bandas
def step_1(matrix_color, matrix_suma):
    #La funciÃ³n gpuarray.if_positive evalua cada posiciÃ³n de la matriz
    #Y de acuerdo a su valor realiza la primer operaciÃ³n o la segunda constatando una sentencia If Else
    matrix_1 = gpuarray.if_positive(matrix_suma,(3*matrix_color)/matrix_suma,matrix_suma)
    return matrix_1

# El step_2 realiza la multiplicaciÃ³n posiciÃ³n a posiciÃ³n del step_1 por la pancromatica
def step_2(matrix_1, matrix_image_pan):
    #La funciÃ³n linalg.mulitply realiza la multiplicaciÃ³n elemento a elemento entre dos matrices
    matrix_2 = linalg.multiply(matrix_1, matrix_image_pan)
    return matrix_2

# El step_3 determina el valor maximo y minimo de la banda resultante de la transformada de brovey
def step_3(matrix_1):
    #La funciÃ³n np.max determina el maximo valor de un array
    #La funciÃ³n np.min determina el minimo valor de un array
    mat_max = np.amax(matrix_1.get())
    mat_min = np.amin(matrix_1.get())
    return mat_max, mat_min

# El step_4 realiza un ajuste de riqueza espectral llamado rescale global min
def step_4(matrix_1, matrix_color, mat_max, mat_min):
    #La funciÃ³n lin_comb invoca el kernel elementwise establecido para llevar a cabo el ajuste por rescale global min
    lin_comb(mat_min, matrix_1, mat_max, matrix_color)
    return matrix_color

# Kernel establecido para llevar a cabo el mÃ©todo de rescale global min
lin_comb = ElementwiseKernel(
        "float a, float *x, float b, float *z",
        "z[i] = ((x[i]-a)*255)/(b-a)",
        "linear_combination")

def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    #Verifica que ambas imagenes cumplan con las condiciones
    end = 0
    start = 0
    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')

    # Convierte a float32 y separa las bandas RGB de la multispectral
    multispectral = multispectral.astype(np.float32)
    r = multispectral[:,:,0].astype(np.float32)
    g = multispectral[:,:,1].astype(np.float32)
    b = multispectral[:,:,2].astype(np.float32)
    b = b.astype(np.float32)
    # Convierte la pancromatica a float32
    panchromatic = panchromatic.astype(np.float32)
    # Suma las bandas de la multispectral
    msuma = r+g+b
    start=time.time()
    r_gpu = gpuarray.to_gpu(r)
    g_gpu = gpuarray.to_gpu(g)
    b_gpu = gpuarray.to_gpu(b)
    panchromatic_gpu = gpuarray.to_gpu(panchromatic)
    msuma_gpu = gpuarray.to_gpu(msuma)

    linalg.init()
    m11_gpu = step_1(r_gpu, msuma_gpu)
    m22_gpu = step_2(m11_gpu, panchromatic_gpu)

    m33_gpu = step_1(b_gpu, msuma_gpu)
    m44_gpu = step_2(m33_gpu, panchromatic_gpu)

    m55_gpu = step_1(g_gpu, msuma_gpu)
    m66_gpu = step_2(m55_gpu, panchromatic_gpu)

    Amax_host, Amin_host = step_3(m22_gpu)
    rr_gpu = gpuarray.empty_like(r_gpu)
    step_4(m22_gpu, rr_gpu, Amax_host, Amin_host)

    Amax_host, Amin_host = step_3(m66_gpu)
    gg_gpu = gpuarray.empty_like(g_gpu)
    step_4(m66_gpu, gg_gpu, Amax_host, Amin_host)

    Amax_host, Amin_host = step_3(m44_gpu)
    bb_gpu = gpuarray.empty_like(b_gpu)
    step_4(m44_gpu, bb_gpu, Amax_host, Amin_host)
    end = time.time()

    ggg_host = gg_gpu.get().astype(np.uint8)
    rrr_host = rr_gpu.get().astype(np.uint8)
    bbb_host = bb_gpu.get().astype(np.uint8)

    # Combina las bandas resultantes
    fusioned_image = np.stack((rrr_host, ggg_host, bbb_host),axis=2)
    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/broveygpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('broveygpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image



# -*- coding: utf-8 -*-
import numpy as np
from numpy import linalg as la
import skimage.io
import numpy as np
import sys
import time

# Funcion para calcular la matriz de varianza-covarianza
def varianza_cov(r_band,g_band, b_band, mat_means, size):
    var_cov_mat = np.zeros((size,size))
    for i in range(r_band.shape[0]):
        for j in range(r_band.shape[1]):
            temp_a = np.array([[r_band[i][j]],[g_band[i][j]],[b_band[i][j]]]-mat_means)
            temp_b = temp_a.transpose()
            mul_temps = np.multiply(temp_a,temp_b)
            var_cov_mat= var_cov_mat+mul_temps
    return var_cov_mat

# Funcion encargada de realizar las potencias sucesivas
def successive_powers(ortogonal_matrix):
    size_mat_ort = len(ortogonal_matrix)
    s = np.zeros((size_mat_ort,1))
    B = np.zeros((size_mat_ort,size_mat_ort))
    for i in range(1, (size_mat_ort+1)):
        B=la.matrix_power(ortogonal_matrix,i)
        s[i-1]=np.trace(B)
    return s

# Esta funcion tiene como objetivo calcular los coeficientes del polinomio caracteristico
def polynomial_coefficients(polynomial_trace, ortogonal_matrix):
    n_interations = len(ortogonal_matrix)
    polynomial = np.zeros((n_interations))
    polynomial[0] = -polynomial_trace[0]
    for i in range(1,n_interations):
        polynomial[i]=-polynomial_trace[i]/(i+1)
        for j in range(i):
            polynomial[i]=polynomial[i]-(polynomial[j]*(polynomial_trace[(i-j)-1])/(i+1))
    return polynomial

# Funcion para calcular los vectores propios normalizados
def eigenvectors_norm(mat_eigenvalues, ortogonal_matrix, mat_eigenvectors):
    n = len(mat_eigenvalues)
    V = np.zeros((n,n))
    S = np.zeros((n,1))
    for i in range(n):
        B= ortogonal_matrix[1:n,1:n]-mat_eigenvalues[i,i]*np.eye(n-1)
        temp_s=la.lstsq(B,mat_eigenvectors,rcond=-1)[0].transpose()
        S=np.insert(temp_s,0,1);
        V[0:n,i]=S/la.norm(S)
    return V, V.transpose()

# Funcion para calcular los componentes principales originales
def componentes_principales_original(r_band, g_band, b_band, mat_q):
    acp = np.zeros((r_band.shape[0],r_band.shape[1]))
    cp1 = np.zeros((r_band.shape[0],r_band.shape[1]))
    cp2 = np.zeros((g_band.shape[0],g_band.shape[1]))
    cp3 = np.zeros((b_band.shape[0],b_band.shape[1]))
    for i in range(r_band.shape[0]):
        for j in range(r_band.shape[1]):
            acp = np.dot(mat_q,np.array(([r_band[i,j]],[g_band[i,j]],[b_band[i,j]])))
            cp1[i,j] = acp[0,0]
            cp2[i,j] = acp[1,0]
            cp3[i,j] = acp[2,0]
    cp2 = (-1)*cp2
    return cp1, cp2, cp3

# Funcion para calcular los componentes principales mediante la pancromatica
def componentes_principales_panchromartic(panchromatic, cp_2, cp_3, inv_eigenvector):
    nb1 = np.zeros((cp_2.shape[0],cp_2.shape[1]))
    nb2 = np.zeros((cp_2.shape[0],cp_2.shape[1]))
    nb3 = np.zeros((cp_2.shape[0],cp_2.shape[1]))
    for i in range(cp_2.shape[0]):
        for j in range(cp_2.shape[1]):
            iacp = np.dot(inv_eigenvector,np.array(([panchromatic[i,j]],[cp_2[i,j]],[cp_3[i,j]])))
            nb1[i,j] = iacp[0,0]
            nb2[i,j] = iacp[1,0]
            nb3[i,j] = iacp[2,0]
    return nb1, nb2, nb3

# Funcion que realiza un ajuste cuando el valor de un pixel es menor a 0
def negative_adjustment(band):
    for i in range(band.shape[0]):
        for j in range(band.shape[1]):
            if band[i,j] < 0:
                band[i,j] = 0.0
    return band

def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    end = 0
    start = 0

    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')

    size_rgb = multispectral.shape


    # Convierte a float32 y separa las bandas RGB de la multispectral
    mmultispectral = multispectral.astype(np.float32)
    r = multispectral[:,:,0].astype(np.float32)
    g = multispectral[:,:,1].astype(np.float32)
    b = multispectral[:,:,2].astype(np.float32)
    # Convierte la pancromatica a float32
    panchromatic = panchromatic.astype(np.float32)


    # Inicial el time_calculated de ejecucion
    start=time.time()

    # Se calcula la media de cada una de las bandas y se forma un arreglo con estos valores
    m1 = np.mean(r)
    m2 = np.mean(g)
    m3 = np.mean(b)
    m_means = np.array([[m1],[m2],[m3]])

    # Se obtiene el numero de bandas
    n_bands = size_rgb[2]

    #Se obtiene la matrix de varianza y covarianza
    mat_var_cov = varianza_cov(r,g,b,m_means,n_bands)

    # Coeficiente para diaganalizar ortogonalmente
    coefficient = 1.0/((size_rgb[0]*size_rgb[1])-1)

    # Matriz diagonalizada ortogonalmente
    ortogonal_matrix = mat_var_cov*coefficient

    # Se calcula la traza de las sucesivas potencias de la matriz ortogonal inicial
    polynomial_trace = successive_powers(ortogonal_matrix)

    # Se calculan los coeficientes del polinomio caracteristico
    characteristic_polynomial = polynomial_coefficients(polynomial_trace, ortogonal_matrix)

    # Se obtienen las raices del polinomio caracteristico
    characteristic_polynomial_roots = np.roots(np.insert(characteristic_polynomial,0,1))

    # Los vectores propios aparecen en la diagonal de la matriz eigenvalues_mat
    eigenvalues_mat = np.diag(characteristic_polynomial_roots)

    # Vectores propios para cada valor propio
    eigenvectors_mat = -1*ortogonal_matrix[1:n_bands,0]

    # Se calcular los vectores propios normalizados
    #Cada vector propio es una columna de la matriz mat_ortogonal_base
    mat_ortogonal_base, q_matrix = eigenvectors_norm(eigenvalues_mat, ortogonal_matrix, eigenvectors_mat)
    w1 = q_matrix[0,:]
    w2 = (-1)*q_matrix[1,:]
    w3 = q_matrix[2,:]
    eigenvectors = np.array((w1,w2,w3))

    # Se calcula la inversa de los vectores propios
    inv_eigenvectors = la.inv(eigenvectors)

    # Se calculan los componentes principales con las bandas originales y los vectores propios
    pc_1, pc_2, pc_3 = componentes_principales_original(r,g,b,q_matrix)

    # Se calculan los componentes con la pancromatica, componentes principales originales 2 y 3, y la inversa de los vectores propios
    nb_1, nb_2, nb_3 = componentes_principales_panchromartic(panchromatic, pc_2, pc_3, inv_eigenvectors)

    nb11 = nb_1.astype(np.float32)
    nb22 = nb_2.astype(np.float32)
    nb33 = nb_3.astype(np.float32)

    # Se realiza un ajuste cuando los valores de cada pixel es menor a 0
    nb11 = negative_adjustment(nb11)
    nb22 = negative_adjustment(nb22)
    nb33 = negative_adjustment(nb33)


    nb111 = nb11.astype(np.uint8)
    nb222 = nb22.astype(np.uint8)
    nb333 = nb33.astype(np.uint8)

    end = time.time()

    # Se realiza un stack con los componentes resultantes
    fusioned_image = np.stack((nb111,nb222,nb333),axis=2)
    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/pcacpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('pcacpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image



# -*- coding: utf-8 -*-
import skimage.io
import numpy as np
import cv2
import time
import sys, getopt

# El step_1 realiza la multiplicaciÃ³n elemento a elemento de una banda con la pancromÃ¡tica
def step_1(color_matrix, image_matrix):
    input_matrix = np.multiply(color_matrix,image_matrix)
    return input_matrix


# El step_2 determina el valor maximo y minimo de la banda resultante de la transformada de brovey
def step_2(matrix_1):
    mat_max = np.amax(matrix_1)
    mat_min = np.amin(matrix_1)
    return mat_max, mat_min


def step_3(matrix_1, mat_max, mat_min):
    matrix_color = np.empty_like(matrix_1)
    for m in range(matrix_1.shape[0]):
        for n in range(matrix_1.shape[0]):
            matrix_color[m,n] = (((matrix_1[m,n]-mat_min)*255)/(mat_max-mat_min))
    return matrix_color

end = 0
start = 0

def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    end = 0
    start = 0
    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')

    # Convierte a float32 y separa las bandas RGB de la multispectral
    multispectral = multispectral.astype(np.float32)
    r = multispectral[:,:,0].astype(np.float32)
    g = multispectral[:,:,1].astype(np.float32)
    b = multispectral[:,:,2].astype(np.float32)
    # Convierte la pancromatica a float32
    panchromatic = panchromatic.astype(np.float32)

    start = time.time()

    m33 = step_1(r, panchromatic)
    m44 = step_1(g, panchromatic)
    m55 = step_1(b, panchromatic)

    Amax, Amin = step_2(m33)
    br = step_3(m33, Amax, Amin)

    Amax, Amin = step_2(m44)
    bg = step_3(m44, Amax, Amin)

    Amax, Amin = step_2(m55)
    bb = step_3(m55, Amax, Amin)

    end = time.time()

    brr = br.astype(np.uint8)
    bgg = bg.astype(np.uint8)
    bbb = bb.astype(np.uint8)

    fusioned_image = np.stack((brr, bgg, bbb),axis=2)
    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/multiplicativecpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('multiplicativecpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image



# -*- coding: utf-8 -*-
import skimage.io
import numpy as np
from numpy import linalg as la
import cv2
import pycuda.autoinit
import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
from pycuda import compiler
import numpy as np
import skcuda.linalg as linalg
import skcuda.misc as misc
from pycuda.elementwise import ElementwiseKernel
import pycuda.cumath as cumath
import sys, getopt
import time


BLOCK_SIZE = 32
# Kernel template para llevar a cabo el calculo de la matriz de varianza-covarianza
kernel_var_cov = """
#include <stdio.h>
__global__ void CovarianceKernel(float *R, float *G, float *B,float *D)
{
    // Thread index
    const uint tx = threadIdx.x;
    const uint ty = threadIdx.y;
    __shared__ float prueba_salida;
    if (threadIdx.x == 0) prueba_salida = 0;
    float valor_temp = 0;
    float salida_temp[9];
    __syncthreads();

    const int size = 3;
    float arreglo[size];

    arreglo[0] = R[ty * %(BLOCK_SIZE)s + tx];
    arreglo[1] = G[ty * %(BLOCK_SIZE)s + tx];
    arreglo[2] = B[ty * %(BLOCK_SIZE)s + tx];
    __syncthreads();

    for(int k = 0; k < 3; k++){
        for(int h = 0; h < 3; h++){
            valor_temp = arreglo[k]*arreglo[h];
            salida_temp[k*3+h] = valor_temp;
            valor_temp = 0;
        }
    }
    __syncthreads();

   for (int i = 0; i < 9; ++i){
    atomicAdd(&prueba_salida,salida_temp[i]);
    __syncthreads();
    D[i] += prueba_salida;
    __syncthreads();
    prueba_salida = 0.0;
    __syncthreads();
   }

}
"""
# Kernel template para llevar a cabo el calculo de componente principales originales
kernel_componentes_principales_original = """
#include <stdio.h>
__global__ void componentesPrincipalesOriginal(float *R, float *G, float *B, float *Q, float *S1, float *S2, float *S3)
{
    // Thread index
    const uint tx = threadIdx.x;
    const uint ty = threadIdx.y;


    const int size = 3;
    float salida_temp [size];
    float valor_temp = 0.0;
    float arreglo[size];
    //__syncthreads();




    arreglo[0] = R[ty * %(BLOCK_SIZE)s + tx];
    arreglo[1] = G[ty * %(BLOCK_SIZE)s + tx];
    arreglo[2] = B[ty * %(BLOCK_SIZE)s + tx];
    __syncthreads();



    for(int i = 0; i < 3; ++i){
        for(int j = 0; j < 3; ++j){
            valor_temp += (Q[i*3+j] * arreglo[j]);
        }
        salida_temp[i] = valor_temp;
        valor_temp = 0.0;
    }
    __syncthreads();

    S1[ty * %(BLOCK_SIZE)s + tx] = salida_temp[0];
    __syncthreads();
    S2[ty * %(BLOCK_SIZE)s + tx] = (-1.0)*salida_temp[1];
    __syncthreads();
    S3[ty * %(BLOCK_SIZE)s + tx] = salida_temp[2];
    __syncthreads();

}
"""

# Kernel template para llevar a cabo el calculo de componente principales con la pancromática
kernel_componentes_principales_pancromatica = """
#include <stdio.h>
__global__ void componentesPrincipalesPancromatica(float *R, float *G, float *B, float *E, float *S1, float *S2, float *S3)
{
    // Thread index
    const uint tx = threadIdx.x;
    const uint ty = threadIdx.y;


    const int size = 3;
    float salida_temp [size];
    float valor_temp = 0.0;
    float arreglo[size];





    arreglo[0] = R[ty * %(BLOCK_SIZE)s + tx];
    arreglo[1] = G[ty * %(BLOCK_SIZE)s + tx];
    arreglo[2] = B[ty * %(BLOCK_SIZE)s + tx];
    __syncthreads();



    for(int i = 0; i < 3; ++i){
        for(int j = 0; j < 3; ++j){
            valor_temp += (E[i*3+j] * arreglo[j]);
        }
        salida_temp[i] = valor_temp;
        valor_temp = 0.0;
    }
    __syncthreads();

    S1[ty * %(BLOCK_SIZE)s + tx] = salida_temp[0];
    __syncthreads();
    S2[ty * %(BLOCK_SIZE)s + tx] = salida_temp[1];
    __syncthreads();
    S3[ty * %(BLOCK_SIZE)s + tx] = salida_temp[2];
    __syncthreads();

}
"""

# Función para dividir una matriz en un submatrices cuadradas
def split(array, nrows, ncols):
    """Split a matrix into sub-matrices."""
    r, h = array.shape
    return (array.reshape(h//nrows, nrows, -1, ncols)
                 .swapaxes(1, 2)
                 .reshape(-1, nrows, ncols))

# Funcion para calcular la matriz de varianza-covarianza
def varianza_cov( R_s, G_s, B_s):
    kernel_code = kernel_var_cov % {
        'BLOCK_SIZE': BLOCK_SIZE
        }
    mod = compiler.SourceModule(kernel_code)
    covariance_kernel = mod.get_function("CovarianceKernel")
    salida_gpu = gpuarray.zeros((3, 3), np.float32)
    Rs_gpu = gpuarray.to_gpu(R_s)
    Gs_gpu = gpuarray.to_gpu(G_s)
    Bs_gpu = gpuarray.to_gpu(B_s)
    for i in range(len(R_s)):
        covariance_kernel(
            # inputs
            Rs_gpu[i], Gs_gpu[i], Bs_gpu[i],
            # output
            salida_gpu,
            # block of multiple threads
            block = (32, 32, 1),
            )
    return salida_gpu.get()

# Función para realizar el Stack de las submatrices resultantes de alguna rutina realizada
def stack_values(list_cp, array_split, size, block_size):
    block_size = block_size
    valor_inicial = 0
    valor_final = 0
    list_cp_nueva = []
    factor_div = (size//block_size)
    factor_ite = len(array_split)//factor_div
    for i in range(factor_ite):
        valor_final = valor_final + factor_div
        list_cp_nueva.append(np.hstack(list_cp[valor_inicial:valor_final]))
        valor_inicial = valor_inicial + factor_div
    cp_final = np.vstack(list_cp_nueva)
    return cp_final

# Funcion para calcular los componentes principales originales
def componentes_principales_original(r_s , g_s, b_s, q, size, block_size):
    cp1_temp = []
    cp2_temp = []
    cp3_temp = []
    size = size
    block_size = block_size
    kernel_code = kernel_componentes_principales_original % {
        'BLOCK_SIZE': BLOCK_SIZE,
        }
    mod = compiler.SourceModule(kernel_code)
    kernel = mod.get_function("componentesPrincipalesOriginal")
    s1_gpu = gpuarray.zeros((block_size,block_size),np.float32)
    s2_gpu = gpuarray.zeros((block_size,block_size),np.float32)
    s3_gpu = gpuarray.zeros((block_size,block_size),np.float32)
    q_gpu = gpuarray.to_gpu(q)
    Rs_gpu_t = gpuarray.to_gpu(r_s)
    Gs_gpu_t = gpuarray.to_gpu(g_s)
    Bs_gpu_t = gpuarray.to_gpu(b_s)
    for i in range(len(r_s)):
        kernel(
            # inputs
            Rs_gpu_t[i], Gs_gpu_t[i], Bs_gpu_t[i], q_gpu,
            # output
            s1_gpu, s2_gpu, s3_gpu,
            # block of multiple threads
            block = (block_size, block_size, 1),
            )
        cp1_temp.append(s1_gpu.get())
        cp2_temp.append(s2_gpu.get())
        cp3_temp.append(s3_gpu.get())
    cp1 = stack_values(cp1_temp, r_s, size, block_size)
    cp2 = stack_values(cp2_temp, r_s, size, block_size)
    cp3 = stack_values(cp3_temp, r_s, size, block_size)
    return cp1, cp2, cp3

# Funcion para calcular los componentes principales mediante la pancromática
def componentes_principales_panchromartic(r_s , g_s, b_s, q, size, block_size):
    block_size = block_size
    nb1_temp = []
    nb2_temp = []
    nb3_temp = []
    size = size
    kernel_code = kernel_componentes_principales_pancromatica % {
        'BLOCK_SIZE': BLOCK_SIZE,
        }
    mod = compiler.SourceModule(kernel_code)
    kernel = mod.get_function("componentesPrincipalesPancromatica")
    s1_gpu = gpuarray.zeros((block_size,block_size),np.float32)
    s2_gpu = gpuarray.zeros((block_size,block_size),np.float32)
    s3_gpu = gpuarray.zeros((block_size,block_size),np.float32)
    Rs_gpu_t = gpuarray.to_gpu(r_s)
    Gs_gpu_t = gpuarray.to_gpu(g_s)
    Bs_gpu_t = gpuarray.to_gpu(b_s)
    q_gpu = gpuarray.to_gpu(q)
    for i in range(len(r_s)):
        kernel(
            # inputs
            Rs_gpu_t[i], Gs_gpu_t[i], Bs_gpu_t[i], q_gpu,
            # output
            s1_gpu, s2_gpu, s3_gpu,
            # block of multiple threads
            block = (block_size, block_size, 1),
            )
        nb1_temp.append(s1_gpu.get())
        nb2_temp.append(s2_gpu.get())
        nb3_temp.append(s3_gpu.get())
    nb1 = stack_values(nb1_temp, g_s, size, block_size)
    nb2 = stack_values(nb2_temp, g_s, size, block_size)
    nb3 = stack_values(nb3_temp, g_s, size, block_size)
    return nb1, nb2, nb3



# Kernel establecido para llevar a cabo la resta de las medias de cada banda a cada pixel de una banda
substract = ElementwiseKernel(
        "float *x, float y, float *z",
        "z[i] = x[i]-y",
        "substract_value")

# Kernel establecido para realiza un ajuste cuando el valor de un pixel es menor a 0
negative_adjustment = ElementwiseKernel(
        "float *x, float *z",
        "if(x[i] < 0){z[i] = 0.0;}else{z[i] = x[i];}",
        "adjust_value")


# Funcion encargada de realizar las potencias sucesivas
def successive_powers(ortogonal_matrix):
    size_mat_ort = len(ortogonal_matrix)
    s = np.zeros((size_mat_ort,1))
    B = np.zeros((size_mat_ort,size_mat_ort))
    for i in range(1, (size_mat_ort+1)):
        B=la.matrix_power(ortogonal_matrix,i)
        s[i-1]=np.trace(B)
    return s

# Esta funcion tiene como objetivo calcular los coeficientes del polinomio caracteristico
def polynomial_coefficients(polynomial_trace, ortogonal_matrix):
    n_interations = len(ortogonal_matrix)
    polynomial = np.zeros((n_interations))
    polynomial[0] = -polynomial_trace[0]
    for i in range(1,n_interations):
        polynomial[i]=-polynomial_trace[i]/(i+1)
        for j in range(i):
            polynomial[i]=polynomial[i]-(polynomial[j]*(polynomial_trace[(i-j)-1])/(i+1))
    return polynomial

# Funcion para calcular los vectores propios normalizados
def eigenvectors_norm(mat_eigenvalues, ortogonal_matrix, mat_eigenvectors):
    n = len(mat_eigenvalues)
    V = np.zeros((n,n))
    S = np.zeros((n,1))
    for i in range(n):
        B= ortogonal_matrix[1:n,1:n]-mat_eigenvalues[i,i]*np.eye(n-1)
        temp_s=la.lstsq(B,mat_eigenvectors,rcond=-1)[0].transpose()
        S=np.insert(temp_s,0,1);
        V[0:n,i]=S/la.norm(S)
    return V, V.transpose()

def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    end = 0
    start = 0

    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')

    size_rgb = multispectral.shape

    # Definición del tamaño del bloque
    BLOCK_SIZE = 32


    # Convierte a float32 y separa las bandas RGB de la multispectral
    m_host = multispectral.astype(np.float32)
    r_host = m_host[:,:,0].astype(np.float32)
    g_host = m_host[:,:,1].astype(np.float32)
    b_host = m_host[:,:,2].astype(np.float32)
    size_rgb = multispectral.shape
    # Convierte la pancromatica a float32
    panchromatic_host = panchromatic.astype(np.float32)


    # Inicial el time_calculated de ejecucion
    start=time.time()

    # Se pasan los array en el host al device
    r_gpu = gpuarray.to_gpu(r_host)
    g_gpu = gpuarray.to_gpu(g_host)
    b_gpu = gpuarray.to_gpu(b_host)
    p_gpu = gpuarray.to_gpu(panchromatic_host)

    # Se calcula la media de cada una de las bandas y se forma un arreglo con estos valores, todo esto en GPU
    mean_r_gpu = misc.mean(r_gpu)
    mean_g_gpu = misc.mean(g_gpu)
    mean_b_gpu = misc.mean(b_gpu)

    # Se obtiene el numero de bandas
    n_bands = size_rgb[2]

    # Se aparta memoria en GPU
    r_gpu_subs = gpuarray.zeros_like(r_gpu,np.float32)
    g_gpu_subs = gpuarray.zeros_like(g_gpu,np.float32)
    b_gpu_subs = gpuarray.zeros_like(b_gpu,np.float32)

    # Se realiza la resta de su respectiva media a cada uno de los pixeles de cada banda,
    substract( r_gpu, mean_r_gpu.get(), r_gpu_subs)
    substract( g_gpu, mean_g_gpu.get(), g_gpu_subs)
    substract( b_gpu, mean_b_gpu.get(), b_gpu_subs)

    # Se divide cada una de las bandas después de ser restada su media, en un conjunto de submatrices cuadradas del tamaño del bloque
    r_subs_split = split(r_gpu_subs.get(),BLOCK_SIZE,BLOCK_SIZE)
    g_subs_split = split(g_gpu_subs.get(),BLOCK_SIZE,BLOCK_SIZE)
    b_subs_split = split(b_gpu_subs.get(),BLOCK_SIZE,BLOCK_SIZE)

    #Se obtiene la matrix de varianza y covarianza
    mat_var_cov = varianza_cov(r_subs_split,g_subs_split,b_subs_split)

    # Coeficiente para diaganalizar ortogonalmente
    coefficient = 1.0/((size_rgb[0]*size_rgb[1])-1)

    # Matriz diagonalizada ortogonalmente
    ortogonal_matrix = mat_var_cov*coefficient

    # Se calcula la traza de las sucesivas potencias de la matriz ortogonal inicial
    polynomial_trace = successive_powers(ortogonal_matrix)


    # Se calculan los coeficientes del polinomio caracteristico
    characteristic_polynomial = polynomial_coefficients(polynomial_trace, ortogonal_matrix)

    # Se obtienen las raices del polinomio caracteristico
    characteristic_polynomial_roots = np.roots(np.insert(characteristic_polynomial,0,1))


    # Los vectores propios aparecen en la diagonal de la matriz eigenvalues_mat
    eigenvalues_mat = np.diag(characteristic_polynomial_roots)


    # Vectores propios para cada valor propio
    eigenvectors_mat = -1*ortogonal_matrix[1:n_bands,0]

    # Se calcular los vectores propios normalizados
    # Cada vector propio es una columna de la matriz mat_ortogonal_base
    mat_ortogonal_base, q_matrix = eigenvectors_norm(eigenvalues_mat, ortogonal_matrix, eigenvectors_mat)
    q_matrix_list = q_matrix.tolist()
    q_matrix_cpu = np.array(q_matrix_list).astype(np.float32)
    w1 = q_matrix_cpu[0,:]
    w2 = (-1)*q_matrix_cpu[1,:]
    w3 = q_matrix_cpu[2,:]
    eigenvectors = np.array((w1,w2,w3))

    # Se calcula la inversa de los vectores propios
    inv_eigenvectors = la.inv(eigenvectors)
    inv_list = inv_eigenvectors.tolist()
    inv_eigenvector_cpu = np.array(inv_list).astype(np.float32)

    # Se realiza la división de las bandas en submatrices del tamaño del bloque
    r_subs_split_cp = split(r_host,BLOCK_SIZE,BLOCK_SIZE)
    g_subs_split_cp = split(g_host,BLOCK_SIZE,BLOCK_SIZE)
    b_subs_split_cp = split(b_host,BLOCK_SIZE,BLOCK_SIZE)

    # Se calculan los componentes principales con las bandas originales y los vectores propios
    pc_1,pc_2,pc_3 = componentes_principales_original(r_subs_split_cp,g_subs_split_cp,b_subs_split_cp,q_matrix_cpu,r_host.shape[0], BLOCK_SIZE)

    # Se realiza la división en submatrices de la pancromática, el componente principal 2 y 3, del tamaño del bloque,
    p_subs_split_nb = split(panchromatic_host,BLOCK_SIZE,BLOCK_SIZE)
    pc_2_subs_split_nb = split(pc_2,BLOCK_SIZE,BLOCK_SIZE)
    pc_3_subs_split_nb = split(pc_3,BLOCK_SIZE,BLOCK_SIZE)

    # Se calculan los componentes con la pancromatica, componentes principales originales 2 y 3, y la inversa de los vectores propios
    nb1,nb2,nb3 = componentes_principales_panchromartic(p_subs_split_nb,pc_2_subs_split_nb,pc_3_subs_split_nb,inv_eigenvector_cpu,r_host.shape[0], BLOCK_SIZE)

    nb11 = nb1.astype(np.float32)
    nb22 = nb2.astype(np.float32)
    nb33 = nb3.astype(np.float32)


    nb11_gpu = gpuarray.to_gpu(nb11)
    nb22_gpu = gpuarray.to_gpu(nb22)
    nb33_gpu = gpuarray.to_gpu(nb33)

    # Se separa espacio en memoria para las matrices resultado de realizar el ajuste
    nb111_gpu = gpuarray.empty_like(nb11_gpu)
    nb222_gpu = gpuarray.empty_like(nb22_gpu)
    nb333_gpu = gpuarray.empty_like(nb33_gpu)

    # Se realiza un ajuste cuando los valores de cada pixel es menor a 0, en GPU
    negative_adjustment(nb11_gpu,nb111_gpu)
    negative_adjustment(nb22_gpu,nb222_gpu)
    negative_adjustment(nb33_gpu,nb333_gpu)

    nb111_cpu = nb111_gpu.get().astype(np.uint8)
    nb222_cpu = nb222_gpu.get().astype(np.uint8)
    nb333_cpu = nb333_gpu.get().astype(np.uint8)


    end = time.time()


    fusioned_image=np.stack((nb111_cpu,nb222_cpu,nb333_cpu),axis=2);
    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/pcagpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('pcagpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image



from setuptools import setup

setup(
    name='sallfus',
    packages=['sallfus', 'sallfus.fusion'], # Mismo nombre que en la estructura de carpetas de arriba
    version='0.1',
    license='LGPL v3', # La licencia que tenga tu paqeute
    description='Specialized library to carry out the fusion of satellite images using different techniques.',
    author='Parall UD',
    author_email='parallud2019@gmail.com',
    url='https://github.com/Parall-UD/sallfus', # Usa la URL del repositorio de GitHub
    download_url='https://github.com/Parall-UD/sallfus/blob/master/dist/sallfus-0.1.tar.gz', # Te lo explico a continuación
    keywords = ['fusion', 'cuda', 'images', 'Satellite', 'quality'], # Palabras que definan tu paquete
    classifiers=['Programming Language :: Python',  # Clasificadores de compatibilidad con versiones de Python para tu paqeute
                 'Programming Language :: Python :: 3.6'],
)



# -*- coding: utf-8 -*-
import skimage.io
from skimage.color import rgb2hsv, hsv2rgb
from scipy import ndimage
import numpy as np
import sys
import time

def adjust_values(matrix):
    for i in range(len(matrix)):
        for j in range(len(matrix)):
            if(matrix[i][j] < 0):
                matrix[i][j] = 0
    return matrix

def hist_match(source, template):
    """
    Adjust the pixel values of a grayscale image such that its histogram
    matches that of a target image

    Arguments:
    -----------
        source: np.ndarray
            Image to transform; the histogram is computed over the flattened
            array
        template: np.ndarray
            Template image; can have different dimensions to source
    Returns:
    -----------
        matched: np.ndarray
            The transformed output image
    """

    oldshape = source.shape
    source = source.ravel()
    template = template.ravel()

    # get the set of unique pixel values and their corresponding indices and
    # counts
    s_values, bin_idx, s_counts = np.unique(source, return_inverse=True,
                                            return_counts=True)
    t_values, t_counts = np.unique(template, return_counts=True)

    # take the cumsum of the counts and normalize by the number of pixels to
    # get the empirical cumulative distribution functions for the source and
    # template images (maps pixel value --> quantile)
    s_quantiles = np.cumsum(s_counts).astype(np.float64)
    s_quantiles /= s_quantiles[-1]
    t_quantiles = np.cumsum(t_counts).astype(np.float64)
    t_quantiles /= t_quantiles[-1]

    # interpolate linearly to find the pixel values in the template image
    # that correspond most closely to the quantiles in the source image
    interp_t_values = np.interp(s_quantiles, t_quantiles, t_values)

    return interp_t_values[bin_idx].reshape(oldshape)



def fusion_images(multispectral, panchromatic, save_image=False, savepath=None, timeCondition=True):
    end = 0
    start = 0
    #Verifica que ambas imagenes cumplan con las condiciones
    if multispectral.shape[2] == 3:
        print('The Multispectral image has '+str(multispectral.shape[2])+' channels and size of '+str(multispectral.shape[0])+'x'+str(multispectral.shape[1]))
    else:
        sys.exit('The first image is not multispectral')

    if len(panchromatic.shape) == 2:
        print(' The Panchromatic image has a size of '+str(panchromatic.shape[0])+'x'+str(panchromatic.shape[1]))
    else:
        sys.exit('The second image is not panchromatic')


    start=time.time()
    hsv = rgb2hsv(multispectral)
    val = hsv[:,:,2]
    sat = hsv[:,:,1]
    mat = hsv[:,:,0]

    pani = hist_match(panchromatic,val)

    s = np.array([[1/256, 1/64, 3/128, 1/64, 1/256],[1/64, 1/16, 3/32, 1/16, 1/64],[3/128, 3/32, 9/64, 3/32, 3/128],
                    [1/64, 1/16, 3/32, 1/16, 1/64],[1/256, 1/64, 3/128, 1/64, 1/256]])
    I1 = ndimage.correlate(pani, s, mode='constant')

    s1 = np.array([[1/256, 0, 1/64, 0, 3/128, 0, 1/64, 0, 1/256],[0, 0, 0, 0, 0, 0, 0, 0, 0],[1/64, 0, 1/16, 0, 3/32, 0, 1/16, 0, 1/64],[0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [3/128, 0, 3/32, 0, 9/64, 0, 3/32, 0, 3/128],[0, 0, 0, 0, 0, 0, 0, 0, 0],[1/64, 0, 1/16, 0, 3/32, 0, 1/16, 0, 1/64], [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [1/256, 0, 1/64, 0, 3/128, 0, 1/64, 0, 1/256]])
    I2 = ndimage.correlate(I1, s1, mode='constant')

    W1=(pani-I1)
    W1 = adjust_values(W1)
    W1 = W1.astype(np.uint8)

    W2=(I1-I2)
    W2 = adjust_values(W2)
    W2 = W2.astype(np.uint8)

    nint=(panchromatic+W1+W2).astype(np.uint8)

    end = time.time()

    n_hsv1 = np.stack((mat, sat, nint),axis=2)
    n_hsv = n_hsv1
    fusioned_image = hsv2rgb(n_hsv).astype(np.uint8)
    if(save_image):
        # Guarda la imagen resultando de acuerdo al tercer parametro establecido en la linea de ejecución del script
        if(savepath != None):
            t = skimage.io.imsave(savepath+'/atrouscpu_image.tif',fusioned_image, plugin='tifffile')
        else:
            t = skimage.io.imsave('atrouscpu_image.tif',fusioned_image, plugin='tifffile')
    #time_calculated de ejecución para la transformada de Brovey en GPU
    time_calculated = (end-start)
    if(timeCondition):
        return {"image": fusioned_image, "time" :  time_calculated}
    else:
        return fusioned_image
